import numpy as np
import vg
from .validation import check_arity


def flip_faces(faces, face_indices_to_flip=()):
    vg.shape.check(locals(), "faces", (-1, -1))
    check_arity(faces)

    flipped = np.copy(faces)
    flipped[face_indices_to_flip] = np.fliplr(faces[face_indices_to_flip])
    return flipped



from .transform_object import Transform


class TransformMixin:
    def transform(self):
        """
        Begin a composite transform operation. After invoking `.transform()`,
        apply transformations, then invoke `.end()` to create a mesh with
        transformed vertices.

        Does not mutate the callee.

        Returns:
            lacecore.Transform: The transform operation.

        Example:
            >>> transformed = (
                mesh.transform()
                .translate(3.0 * vg.basis.x)
                .uniform_scale(3.0)
                .end()
            )
        """
        return Transform(self)

    def uniformly_scaled(self, factor):
        """
        Scale by the given factor.

        Args:
            factor (float): The scale factor.

        Returns:
            lacecore.Mesh: A mesh with transformed vertices.
        """
        return self.transform().uniform_scale(factor=factor).end()

    def units_converted(self, from_units, to_units):
        """
        Convert the mesh from one set of units to another.

        Support the length units from Ounce:
        https://github.com/lace/ounce/blob/master/ounce/core.py#L26

        Returns:
            lacecore.Mesh: A mesh with transformed vertices.
        """
        return (
            self.transform()
            .convert_units(from_units=from_units, to_units=to_units)
            .end()
        )

    def non_uniformly_scaled(self, x_factor, y_factor, z_factor):
        """
        Scale along each axis by the given factors.

        Args:
            x_factor (flot): The scale factor along the `x` axis.
            y_factor (flot): The scale factor along the `y` axis.
            z_factor (flot): The scale factor along the `z` axis.

        Returns:
            lacecore.Mesh: A mesh with transformed vertices.
        """
        return self.transform().non_uniform_scale(x_factor, y_factor, z_factor).end()

    def faces_flipped(self):
        """
        Flip the orientation of the faces.

        Returns:
            lacecore.Mesh: A mesh with transformed faces.
        """
        return self.transform().flip_faces().end()

    def flipped(self, dim, preserve_vertex_centroid=False):
        """
        Flip about the given axis.

        Args:
            dim (int): The axis to flip around: 0 for `x`, 1 for `y`, 2 for `z`.
            preserve_vertex_centroid (bool): When `True`, translate after
                flipping to preserve the original vertex centroid.

        Returns:
            lacecore.Mesh: A mesh with transformed vertices.
        """
        return (
            self.transform()
            .flip(dim, preserve_vertex_centroid=preserve_vertex_centroid)
            .end()
        )

    def translated(self, translation):
        """
        Translate by the vector provided.

        Args:
            vector (np.arraylike): A 3x1 vector.

        Returns:
            lacecore.Mesh: A mesh with transformed vertices.
        """
        return self.transform().translate(translation=translation).end()

    def reoriented(self, up, look):
        """
        Reorient using up and look.

        Returns:
            lacecore.Mesh: A mesh with transformed vertices.
        """
        return self.transform().reorient(up=up, look=look).end()

    def rotated(self, rotation):
        """
        Rotate by the given 3x3 rotation matrix or a Rodrigues vector.

        Returns:
            lacecore.Mesh: A mesh with transformed vertices.
        """
        return self.transform().rotate(rotation=rotation).end()



from polliwog import Box
import vg


class AnalysisMixin:
    @property
    def vertex_centroid(self):
        """
        The centroid or geometric average of the vertices.
        """
        return vg.average(self.v)

    @property
    def bounding_box(self):
        """
        A bounding box around the vertices.

        Returns:
            polliwog.Box: The bounding box.

        See also:
            https://polliwog.readthedocs.io/en/latest/#polliwog.Box
        """
        return Box.from_points(self.v)

    def apex(self, along):
        """
        Find the most extreme vertex in the direction provided.

        Args:
            along (np.arraylike): A `(3,)` direction of interest.

        Returns:
            np.ndarray: A copy of the point in `self.v` which lies furthest
                in the direction of interest.
        """
        return vg.apex(self.v, along=along)



from collections import OrderedDict
import numpy as np

from ._group_map import GroupMap
from ._mesh import Mesh

ERROR_MESSAGE = "tinyobjloader library has not been installed. \
You will not be able to load OBJ files. \
To fix, run `pip install lacecore[obj]`"

try:
    from tinyobjloader import ObjReader, ObjReaderConfig
except Exception:  # pragma: no cover
    ObjReader = None
    ObjReaderConfig = None


class LoadException(Exception):
    pass


class ArityException(Exception):
    pass


def load(mesh_path, triangulate=False):
    """
    Load a `Mesh` from a path to an OBJ file.

    Args:
        mesh_path (str): A path to an OBJ file
        triangulate (bool): A flag that indicates whether to triangulate the mesh on load.

    Returns:
        lacecore.Mesh: A `Mesh` instance
    """
    if ObjReader is None:  # pragma: no cover
        raise ImportError(ERROR_MESSAGE)
    reader = ObjReader()
    config = ObjReaderConfig()
    # There is some complex code in tinyobjloader which occasionally switches
    # the axes of triangulation based on the vertex positions. This is
    # undesirable in lacecore as it scrambles correspondence.
    config.triangulate = False
    success = reader.ParseFromFile(mesh_path, config)
    if not success:
        raise LoadException(reader.Warning() or reader.Error())
    attrib = reader.GetAttrib()
    shapes = reader.GetShapes()
    tinyobj_vertices = attrib.numpy_vertices().reshape(-1, 3)
    all_vertices_per_face = np.concatenate(
        [shape.mesh.numpy_num_face_vertices() for shape in shapes]
    )
    first_arity = all_vertices_per_face[0]
    if np.any(all_vertices_per_face != first_arity) or first_arity not in (3, 4):
        raise ArityException(
            "OBJ Loader does not support mixed arities, or arities greater than 4 or less than 3"
        )

    segm = OrderedDict()
    all_faces = None

    for shape in shapes:
        tinyobj_all_indices = shape.mesh.numpy_indices().reshape(-1, 3)[:, 0]
        faces = tinyobj_all_indices.reshape(-1, first_arity)
        if triangulate and first_arity == 4:
            # Triangulate ABCD as ABC + ACD.
            faces = faces[:, [[0, 1, 2], [0, 2, 3]]].reshape(-1, 3)
        start = len(all_faces) if all_faces is not None else 0
        end = start + len(faces)
        all_faces = faces if all_faces is None else np.concatenate((all_faces, faces))

        group_names = shape.name.split()
        for name in group_names:
            if name not in segm:
                segm[name] = []
            segm[name] = segm[name] + list(range(start, end))

    group_map = GroupMap.from_dict(segm, len(all_faces))
    return Mesh(v=tinyobj_vertices, f=all_faces, face_groups=group_map)



"""
Functions for creating meshes for tesselated 3D shapes.

See also:
    https://en.wikipedia.org/wiki/Tessellation_(computer_graphics)
"""

from . import _shapes
from ._shapes import *  # noqa: F401,F403

__all__ = _shapes.__all__



import numpy as np
from polliwog import Plane
import vg
from .reconcile_selection import reconcile_selection
from .._common.reindexing import create_submesh
from .._common.validation import check_indices


class Selection:
    """
    Encapsulate a chained submesh selection operation.

    Invoke `.end()` to apply the selection operation and create a submesh. By
    default, orphaned vertices are pruned. However you can keep them by
    invoking `.end(prune_orphan_vertices=True)`.

    Include `.union()` in the chain to combine more than one set of selection
    criteria into a single submesh.

    Args:
        target (lacecore.Mesh): The mesh on which to operate.
        union_with (lacecore.Selection): The operation with which the new
            instance should combine itself. Normally this is reserved for
            internal use.
    """

    def __init__(
        self, target, union_with=[],
    ):
        self._target = target
        self._union_with = union_with
        self._vertex_mask = np.ones(target.num_v, dtype=np.bool)
        self._face_mask = np.ones(target.num_f, dtype=np.bool)

    def _keep_faces(self, mask):
        self._face_mask = np.logical_and(self._face_mask, mask)

    def _keep_vertices(self, mask):
        self._vertex_mask = np.logical_and(self._vertex_mask, mask)

    def vertices_at_or_above(self, dim, point):
        """
        Select vertices which, when projected to the given axis, are either
        coincident with the projection of the given point, or lie further
        along the axis.

        Args:
            dim (int): The axis of interest: 0 for `x`, 1 for `y`, 2 for `z`.
            point (np.arraylike): The point of interest.

        Returns:
            self
        """
        if dim not in [0, 1, 2]:
            raise ValueError("Expected dim to be 0, 1, or 2")
        vg.shape.check(locals(), "point", (3,))
        self._keep_vertices(self._target.v[:, dim] >= point[dim])
        return self

    def vertices_above(self, dim, point):
        """
        Select vertices which, when projected to the given axis, lie further
        along that axis than the projection of the given point.

        Args:
            dim (int): The axis of interest: 0 for `x`, 1 for `y`, 2 for `z`.
            point (np.arraylike): The point of interest.

        Returns:
            self
        """
        if dim not in [0, 1, 2]:
            raise ValueError("Expected dim to be 0, 1, or 2")
        vg.shape.check(locals(), "point", (3,))
        self._keep_vertices(self._target.v[:, dim] > point[dim])
        return self

    def vertices_at_or_below(self, dim, point):
        """
        Select vertices which, when projected to the given axis, are either
        coincident with the projection of the given point, or lie before it.

        Args:
            dim (int): The axis of interest: 0 for `x`, 1 for `y`, 2 for `z`.
            point (np.arraylike): The point of interest.

        Returns:
            self
        """
        if dim not in [0, 1, 2]:
            raise ValueError("Expected dim to be 0, 1, or 2")
        vg.shape.check(locals(), "point", (3,))
        self._keep_vertices(self._target.v[:, dim] <= point[dim])
        return self

    def vertices_below(self, dim, point):
        """
        Select vertices which, when projected to the given axis, lie before
        the projection of the given point.

        Args:
            dim (int): The axis of interest: 0 for `x`, 1 for `y`, 2 for `z`.
            point (np.arraylike): The point of interest.

        Returns:
            self
        """
        if dim not in [0, 1, 2]:
            raise ValueError("Expected dim to be 0, 1, or 2")
        vg.shape.check(locals(), "point", (3,))
        self._keep_vertices(self._target.v[:, dim] < point[dim])
        return self

    def vertices_on_or_in_front_of_plane(self, plane):
        """
        Select the vertices which are either on or in front of the given
        plane.

        Args:
            plane (polliwog.Plane): The plane of interest.

        Returns:
            self

        See also:
            https://polliwog.readthedocs.io/en/latest/#polliwog.Plane
        """
        if not isinstance(plane, Plane):
            raise ValueError("Expected an instance of polliwog.Plane")
        self._keep_vertices(plane.sign(self._target.v) != -1)
        return self

    def vertices_in_front_of_plane(self, plane):
        """
        Select the vertices which are in front of the given plane.

        Args:
            plane (polliwog.Plane): The plane of interest.

        Returns:
            self

        See also:
            https://polliwog.readthedocs.io/en/latest/#polliwog.Plane
        """
        if not isinstance(plane, Plane):
            raise ValueError("Expected an instance of polliwog.Plane")
        self._keep_vertices(plane.sign(self._target.v) == 1)
        return self

    def vertices_on_or_behind_plane(self, plane):
        """
        Select the vertices which are either on or behind the given plane.

        Args:
            plane (polliwog.Plane): The plane of interest.

        Returns:
            self

        See also:
            https://polliwog.readthedocs.io/en/latest/#polliwog.Plane
        """
        if not isinstance(plane, Plane):
            raise ValueError("Expected an instance of polliwog.Plane")
        self._keep_vertices(plane.sign(self._target.v) != 1)
        return self

    def vertices_behind_plane(self, plane):
        """
        Select the vertices which are behind the given plane.

        Args:
            plane (polliwog.Plane): The plane of interest.

        Returns:
            self

        See also:
            https://polliwog.readthedocs.io/en/latest/#polliwog.Plane
        """
        if not isinstance(plane, Plane):
            raise ValueError("Expected an instance of polliwog.Plane")
        self._keep_vertices(plane.sign(self._target.v) == -1)
        return self

    @staticmethod
    def _mask_like(value, num_elements):
        value = np.asarray(value)
        if value.dtype == np.bool:
            vg.shape.check(locals(), "value", (num_elements,))
            return value
        else:
            check_indices(value, num_elements, "mask")
            mask = np.zeros(num_elements, dtype=np.bool)
            mask[value] = True
            return mask

    def pick_vertices(self, indices_or_boolean_mask):
        """
        Select only the given vertices.

        Args:
            indices_or_boolean_mask (np.arraylike): Either a list of vertex
                indices, or a boolean mask the same length as the vertex array.

        Returns:
            self
        """
        self._keep_vertices(
            self._mask_like(indices_or_boolean_mask, len(self._vertex_mask))
        )
        return self

    def pick_faces(self, indices_or_boolean_mask):
        """
        Select only the given faces.

        Args:
            indices_or_boolean_mask (np.arraylike): Either a list of face
                indices, or a boolean mask the same length as the face array.

        Returns:
            self
        """
        self._keep_faces(self._mask_like(indices_or_boolean_mask, len(self._face_mask)))
        return self

    def pick_face_groups(self, *group_names):
        """
        Select faces which belong to the given face groups.

        Args:
            group_names (list): The face groups to keep.

        Returns:
            self
        """
        self._keep_faces(self._target.face_groups.union(*group_names))
        return self

    def union(self):
        """
        Chain on a new selection object. This works like a boolean "or" to
        combine two sets of submesh operations.

        Args:
            indices_or_boolean_mask (np.arraylike): Either a list of face
                indices, or a boolean mask the same length as the face array.

        Returns:
            lacecore.Selection: The new selection operation, which will
                combine itself with `self`.

        Example:
            >>> upper_half_plus_right_half = (
                mesh.select()
                .vertices_above(centroid, dim=0)
                .union()
                .vertices_above(centroid, dim=1)
                .end()
            )

        """
        return self.__class__(target=self._target, union_with=self._union_with + [self])

    def _reconciled_selection(self, prune_orphan_vertices):
        return reconcile_selection(
            faces=self._target.f,
            face_mask=self._face_mask,
            vertex_mask=self._vertex_mask,
            prune_orphan_vertices=prune_orphan_vertices,
        )

    def end(
        self,
        prune_orphan_vertices=True,
        ret_indices_of_original_faces_and_vertices=False,
    ):
        """
        Apply the selection to construct a submesh.

        Args:
            prune_orphan_vertices (bool): When `True`, remove vertices which
                are referenced only by faces which are being removed.
            ret_indices_of_original_faces_and_vertices: When `True`, also
                return the indices of the original faces and vertices.

        Returns:
            object: Either the submesh as an instance of `lacecore.Mesh`, or a tuple
                `(submesh, indices_of_original_faces, indices_of_original_vertices)`.
                The index arrays contain the new indices of the original vertices,
                and `-1` for each removed face and vertex.
        """
        # The approach here is designed to keep faces which have verts in two
        # halves of a union, and to avoid keeping the entire mesh when faces
        # are selected in one half of a union and verts are selected in the
        # other.

        # First, form the union of reconciled vertices.
        _, initial_vertex_mask_of_union = self._reconciled_selection(
            prune_orphan_vertices=prune_orphan_vertices
        )
        for other in self._union_with:
            _, this_vertex_mask = other._reconciled_selection(
                prune_orphan_vertices=prune_orphan_vertices
            )
            initial_vertex_mask_of_union = np.logical_or(
                initial_vertex_mask_of_union, this_vertex_mask
            )

        # Second, union the faces.
        initial_face_mask_of_union = self._face_mask
        for other in self._union_with:
            initial_face_mask_of_union = np.logical_or(
                initial_face_mask_of_union, other._face_mask
            )

        # Finally, reconcile the union of reconciled vertices with the union of
        # faces.
        face_mask_of_union, vertex_mask_of_union = reconcile_selection(
            faces=self._target.f,
            face_mask=initial_face_mask_of_union,
            vertex_mask=initial_vertex_mask_of_union,
            prune_orphan_vertices=prune_orphan_vertices,
        )

        return create_submesh(
            mesh=self._target,
            vertex_mask=vertex_mask_of_union,
            face_mask=face_mask_of_union,
            ret_indices_of_original_faces_and_vertices=ret_indices_of_original_faces_and_vertices,
        )



import numpy as np


def indices_of_original_elements_after_applying_mask(mask):
    """
    Given a mask that represents which of the original elements should be kept,
    produce an array containing the new indices of the original elements. Returns
    -1 as the index of the removed elements.
    """
    result = np.repeat(np.int(-1), len(mask))
    result[mask] = np.arange(np.count_nonzero(mask))
    return result


def create_submesh(
    mesh, vertex_mask, face_mask, ret_indices_of_original_faces_and_vertices=False
):
    """
    Apply the requested mask to the vertices and faces to create a submesh,
    discarding the face groups.
    """
    from .._mesh import Mesh

    new_v = mesh.v[vertex_mask]
    indices_of_original_vertices = indices_of_original_elements_after_applying_mask(
        vertex_mask
    )
    new_f = indices_of_original_vertices[mesh.f[face_mask]]
    submesh = Mesh(v=new_v, f=new_f)

    if ret_indices_of_original_faces_and_vertices:
        indices_of_original_faces = indices_of_original_elements_after_applying_mask(
            face_mask
        )
        return submesh, indices_of_original_faces, indices_of_original_vertices
    else:
        return submesh



import numpy as np
from polliwog import CompositeTransform
from .._common.tri import flip_faces


class Transform:
    """
    Encapsulate a composite transform operation.

    Invoke `.end()` to apply the transform operation and create a mesh with
    transformed vertices and faces.

    Args:
        target (lacecore.Mesh): The mesh on which to operate.

    See also:
        https://polliwog.readthedocs.io/en/latest/#polliwog.CompositeTransform
    """

    def __init__(self, target):
        self.target = target
        self._transform = CompositeTransform()
        self._flip_faces = False

    def append_transform(self, forward, reverse=None):
        """
        Append an arbitrary transformation, defined by 4x4 forward and reverse
        matrices.

        Returns:
            self
        """
        self._transform.append_transform(forward=forward, reverse=reverse)
        return self

    def uniform_scale(self, factor):
        """
        Scale by the given factor.

        Args:
            factor (float): The scale factor.

        Returns:
            self
        """
        self._transform.uniform_scale(factor=factor)
        return self

    def non_uniform_scale(self, x_factor, y_factor, z_factor):
        """
        Scale along each axis by the given factors.

        Args:
            x_factor (flot): The scale factor along the `x` axis.
            y_factor (flot): The scale factor along the `y` axis.
            z_factor (flot): The scale factor along the `z` axis.

        Returns:
            self
        """
        self._transform.non_uniform_scale(x_factor, y_factor, z_factor)
        return self

    def convert_units(self, from_units, to_units):
        """
        Convert the mesh from one set of units to another.

        Supports the length units from Ounce:
        https://github.com/lace/ounce/blob/master/ounce/core.py#L26

        Returns:
            self
        """
        self._transform.convert_units(from_units=from_units, to_units=to_units)
        return self

    def translate(self, translation):
        """
        Translate by the vector provided.

        Args:
            vector (np.arraylike): A 3x1 vector.

        Returns:
            self
        """
        self._transform.translate(translation=translation)
        return self

    def reorient(self, up, look):
        """
        Reorient using up and look.

        Returns:
            self
        """
        self._transform.reorient(up=up, look=look)
        return self

    def rotate(self, rotation):
        """
        Rotate by the given 3x3 rotation matrix or a Rodrigues vector.

        Returns:
            self
        """
        self._transform.rotate(rotation=rotation)
        return self

    def flip_faces(self):
        """
        Flip the orientation of the faces.

        Returns:
            self
        """
        self._flip_faces = not self._flip_faces
        return self

    def flip(self, dim, preserve_vertex_centroid=False):
        """
        Flip about the given axis.

        Args:
            dim (int): The axis to flip around: 0 for `x`, 1 for `y`, 2 for `z`.
            preserve_vertex_centroid (bool): When `True`, translate after
                flipping to preserve the original vertex centroid.

        Returns:
            self
        """
        if dim not in (0, 1, 2):
            raise ValueError("Expected dim to be 0, 1, or 2")

        self._transform.flip(dim=dim)
        self.flip_faces()

        if preserve_vertex_centroid:
            translation = np.zeros(3)
            # `2 * vertex_centroid[dim]` takes it all the way back to the other
            # side of the axis.
            translation[dim] = 2 * self.target.vertex_centroid[dim]
            self._transform.translate(translation=translation)

        return self

    def end(self, reverse=False):
        """
        Apply the requested transformation and return a new mesh.

        Args:
            reverse (bool): When `True` applies the selected transformations
                in reverse.

        Returns:
            lacecore.Mesh: The transformed mesh.
        """
        from .._mesh import Mesh  # Avoid circular import.

        return Mesh(
            v=self._transform(self.target.v),
            f=flip_faces(self.target.f) if self._flip_faces else self.target.f,
            face_groups=self.target.face_groups,
        )



from ._group_map import GroupMap  # noqa: F401
from ._mesh import Mesh  # noqa: F401
from ._obj import ArityException, LoadException, load as load_obj  # noqa: F401
from ._selection.selection_object import Selection  # noqa: F401
from ._transform.transform_object import Transform  # noqa: F401
from .package_version import __version__  # noqa: F401



from .selection_object import Selection


class SelectionMixin:
    def select(self):
        """
        Begin a chained selection operation. After invoking `.select()`,
        apply selection criteria, then invoke `.end()` to create a submesh.

        Include `.union()` in the chain to combine multiple sets of
        selection criteria into a single submesh.

        Does not mutate the callee.

        Returns:
            lacecore.Selection: The selection operation.

        Example:
            >>> centroid = np.average(mesh.v, axis=0)
            >>> upper_right_quadrant = (
                mesh.select()
                .vertices_above(centroid, dim=0)
                .vertices_above(centroid, dim=1)
                .end()
            )
            >>> upper_half_plus_right_half = (
                mesh.select()
                .vertices_above(centroid, dim=0)
                .union()
                .vertices_above(centroid, dim=1)
                .end()
            )
        """
        return Selection(target=self)

    def keeping_vertices_at_or_above(self, dim, point):
        """
        Select vertices which, when projected to the given axis, are either
        coincident with the projection of the given point, or lie further
        along the axis.

        Return a new mesh, without mutating the callee.

        Args:
            dim (int): The axis of interest: 0 for `x`, 1 for `y`, 2 for `z`.
            point (np.arraylike): The point of interest.

        Returns:
            lacecore.Mesh: A submesh containing the selection.
        """
        return self.select().vertices_at_or_above(dim=dim, point=point).end()

    def keeping_vertices_above(self, dim, point):
        """
        Select vertices which, when projected to the given axis, lie further
        along that axis than the projection of the given point.

        Return a new mesh, without mutating the callee.

        Args:
            dim (int): The axis of interest: 0 for `x`, 1 for `y`, 2 for `z`.
            point (np.arraylike): The point of interest.

        Returns:
            lacecore.Mesh: A submesh containing the selection.
        """
        return self.select().vertices_above(dim=dim, point=point).end()

    def keeping_vertices_at_or_below(self, dim, point):
        """
        Select vertices which, when projected to the given axis, are either
        coincident with the projection of the given point, or lie before it.

        Return a new mesh, without mutating the callee.

        Args:
            dim (int): The axis of interest: 0 for `x`, 1 for `y`, 2 for `z`.
            point (np.arraylike): The point of interest.

        Returns:
            lacecore.Mesh: A submesh containing the selection.
        """
        return self.select().vertices_at_or_below(dim=dim, point=point).end()

    def keeping_vertices_below(self, dim, point):
        """
        Select vertices which, when projected to the given axis, lie before
        the projection of the given point.

        Return a new mesh, without mutating the callee.

        Args:
            dim (int): The axis of interest: 0 for `x`, 1 for `y`, 2 for `z`.
            point (np.arraylike): The point of interest.

        Returns:
            lacecore.Mesh: A submesh containing the selection.
        """
        return self.select().vertices_below(dim=dim, point=point).end()

    def keeping_vertices_on_or_in_front_of_plane(self, plane):
        """
        Select the vertices which are either on or in front of the given
        plane.

        Return a new mesh, without mutating the callee.

        Args:
            plane (polliwog.Plane): The plane of interest.

        Returns:
            lacecore.Mesh: A submesh containing the selection.

        See also:
            https://polliwog.readthedocs.io/en/latest/#polliwog.Plane
        """
        return self.select().vertices_on_or_in_front_of_plane(plane=plane).end()

    def keeping_vertices_in_front_of_plane(self, plane):
        """
        Select the vertices which are in front of the given plane.

        Return a new mesh, without mutating the callee.

        Args:
            plane (polliwog.Plane): The plane of interest.

        Returns:
            lacecore.Mesh: A submesh containing the selection.

        See also:
            https://polliwog.readthedocs.io/en/latest/#polliwog.Plane
        """
        return self.select().vertices_in_front_of_plane(plane=plane).end()

    def keeping_vertices_on_or_behind_plane(self, plane):
        """
        Select the vertices which are either on or behind the given plane.

        Return a new mesh, without mutating the callee.

        Args:
            plane (polliwog.Plane): The plane of interest.

        Returns:
            lacecore.Mesh: A submesh containing the selection.

        See also:
            https://polliwog.readthedocs.io/en/latest/#polliwog.Plane
        """
        return self.select().vertices_on_or_behind_plane(plane=plane).end()

    def keeping_vertices_behind_plane(self, plane):
        """
        Select the vertices which are behind the given plane.

        Return a new mesh, without mutating the callee.

        Args:
            plane (polliwog.Plane): The plane of interest.

        Returns:
            lacecore.Mesh: A submesh containing the selection.

        See also:
            https://polliwog.readthedocs.io/en/latest/#polliwog.Plane
        """
        return self.select().vertices_behind_plane(plane=plane).end()

    def picking_vertices(self, indices_or_boolean_mask):
        """
        Select only the given vertices.

        Return a new mesh, without mutating the callee.

        Args:
            indices_or_boolean_mask (np.arraylike): Either a list of vertex
                indices, or a boolean mask the same length as the vertex array.

        Returns:
            lacecore.Mesh: A submesh containing the selection.
        """
        return (
            self.select()
            .pick_vertices(indices_or_boolean_mask=indices_or_boolean_mask)
            .end()
        )

    def picking_faces(self, indices_or_boolean_mask):
        """
        Select only the given faces.

        Return a new mesh, without mutating the callee.

        Args:
            indices_or_boolean_mask (np.arraylike): Either a list of vertex
                indices, or a boolean mask the same length as the vertex array.

        Returns:
            lacecore.Mesh: A submesh containing the selection.
        """
        return (
            self.select()
            .pick_faces(indices_or_boolean_mask=indices_or_boolean_mask)
            .end()
        )

    def picking_face_groups(self, *group_names):
        """
        Select faces which belong to the given face groups.

        Args:
            group_names (list): The face groups to keep.

        Returns:
            lacecore.Mesh: A submesh containing the selection.
        """
        return self.select().pick_face_groups(*group_names).end()
